#! perl

# --8<--8<--8<--8<--
#
# Copyright (C) 2011 Smithsonian Astrophysical Observatory
#
# This file is part of write-ecf-fits
#
# write-ecf-fits is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# -->8-->8-->8-->8--

use 5.10.0;

use strict;
use warnings;

package WECF;

our $VERSION = '2.0.0';

use Moo;

use Util::Any -list => [ 'mapp', 'uniq' ];

use Params::Validate qw[ validate SCALAR HASHREF ];
use PDL::Primitive qw[ where ];
use PDL::IO::RDB qw[ rdrdb ];

our ( $a, $b );

use RDB;

has energy   => ( is => 'rw', init_arg => undef );
has theta    => ( is => 'rw', init_arg => undef );
has phi      => ( is => 'rw', init_arg => undef );
has fraction => ( is => 'rw', init_arg => undef,
		  isa => sub { $_[0]->isa('PDL') }
		);
has data     => ( is => 'rw', init_arg => undef,
		  isa => sub { 'HASH' eq ref $_[0] }
		);

# read the ECF's from an RDB table
sub BUILD
{
    my $self = shift;

    my %args = validate( @_, {
			      param   => { type => HASHREF },
			      file    => { type => SCALAR },
			      verbose => { type => SCALAR },
			      frac    => { type => SCALAR },
			      epars   => { isa => 'WECF::Pars' },
			      opars   => { isa => 'WECF::OutputParams' },
			     }
		       );

    my %point = %{$args{param}};

    print( "$args{file}: ", join( '; ', mapp { "$a: $b" } %point ), "\n" )
      if $args{verbose} > 1;

    my @cols = ( $args{frac}, $args{epars}->cols );
    my @rcols = @cols;

  {
      my $rdb = RDB->new( $args{file} )
	or die( "$args{file}: error opening/parsing\n" );

      push @rcols, 'real_fit'
	if defined $rdb->defn( 'real_fit' );
  }

    my %data;
    @data{ @rcols } = rdrdb( $args{file}, @rcols );

    # filter data if required
    @data{@cols} = where( @data{@cols}, $data{real_fit} == 2 )
      if exists $data{real_fit};

    # sort data by ascending ecf
    my $idx = $data{$args{frac}}->qsorti;
    @data{@cols} = map { $_->index($idx)->sever } @data{@cols};

    # worry about linked statistics columns
    my %links;
  {
      my $stats = $args{opars}->group('stats');

      my %link;
      %link = $stats->mapvals( points => \%point, flatten => 1 )
	unless ! defined $stats || $stats->is_identity;

      # if this is a self-referential link it's not really a link
      %link = () if %link && grep { $link{$_} == $point{$_} } keys %point;

      $links{stats} = \%link;
  }

    my %idata = 
      $args{epars}->populate(
			     data => \%data,
			     links => \%links,
			     is_linked_to => $args{opars}->is_linked_to( points => \%point 
								       )->sclr,
			    );

    $self->data( \%idata );
    $self->fraction( $data{$args{frac}} );
    $self->energy( $point{energy} );
    $self->theta( $point{theta} );
    $self->phi( $point{phi} );

    return;
}



1;
