#! perl

# --8<--8<--8<--8<--
#
# Copyright (C) 2011 Smithsonian Astrophysical Observatory
#
# This file is part of write-ecf-fits
#
# write-ecf-fits is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# -->8-->8-->8-->8--

use 5.10.0;

use strict;
use warnings;

use Getopt::ArgvFile;

package WECF::Options;

our $VERSION = '2.0.0';

use File::Basename;
use Regexp::Common;
use Date::Format qw[ time2str ];
use Getopt::Long::Descriptive;

use Carp;

use WECF::Pars;

# rather raw delegation to GLD opt or Pars objects object
our $AUTOLOAD;
sub AUTOLOAD {
    $AUTOLOAD =~ s/.*:://;
    my $self = shift;

    my $method = $self->opt && $self->opt->can($AUTOLOAD)
      or die( __PACKAGE__, "::$AUTOLOAD: unknown method\n" );

    unshift @_, $self->opt;
    goto &$method
}

sub DESTROY {}

use Moo;

has re    => ( is => 'ro', init_arg => undef,
	       writer => '_re' );
has opt   => ( is => 'rw', init_arg => undef );
has files => ( is => 'rw', init_arg => undef );
has pars  => ( is => 'rw',
	       isa => sub { $_[0]->isa('WECF::Pars') },
	       init_arg => undef,
	     );
has createdate => ( is => 'ro',
		    init_arg => undef,
		    writer => '_createdate' );

has dontsave => ( is => 'ro',
		  init_arg => undef,
		  default => sub { 
				 return { map { $_ => 1 }
					  qw( help verbose usage rpar fits_cols npar
					      fits_units dontsave createdate version )
					}
			       },
		);

sub BUILD {

    my $self = shift;

    # work around Getopt::Long::Descriptive's not handling
    # --help/--usage/--version well if there's a required option

    my @required = qw[ output detector config output_params ];

    my ( $opt, $usage);
  {
      local $SIG{__WARN__} = sub { die @_ };

      ( $opt, $usage ) = describe_options
	(
	 'usage: ' . basename( $0 ) . ' %o',
	 [ 'Input/Output:' ],
	 [ 'extract_re=s', 'regexp to extract parameters from filename',
	 {
	  default => '(?<energy>$RE{num}{real})[_/](?<theta>$RE{num}{real})[_/](?<phi>$RE{num}{real})'
	 },
	 ],
	 [ 'frac=s',
	   'column containing fractions', { default => 'fraction' } ],

	 [ 'par=s@',
	   'parameter description. repeatable', { default => [] } ],
	 [ 'output_params=s',
	   'file containg list of output values of independent params' ],
	 [ 'op_filter=s',
	   'output parameter filter' ],
	 [ 'err_extra_params',
	   'input files with params not in output list are fatal' ],
	 [ 'output=s',   'output FITS file' ],
	 [ 'denergy=f',  'halfwidth of energy bin', { default => 0.5 } ],
	 [ 'flip_phi',   'add 180 to phi', { default => 0 } ],

	 [ 'FITS Header:' ],
	 [ 'detector=s', 'detector name' ],
	 [ 'config=s',   'raytrace configuration name' ],
	 [ 'comment=s',  'file containing extra comments' ],
	 [ 'vdate=s',    'valid starting date' ],

	 [ 'Misc:' ],
	 [ 'verbose|v+', 'be verbose', { default => 0 } ],
	 [ 'version',    'print version and exit', { default => 0 } ],
	 [ 'help',       'print short help and exit', { default => 0 } ],
	 [ 'usage',	 ' print longer help and exit', { default => 0 } ],
	);
  }

    $self->opt( $opt );

    return if $opt->help || $opt->usage || $opt->version;

    my @notset = grep { ! defined $opt->{$_} } @required;
    die( "required parameter(s) not specified: ",
	 join( ', ', @notset), "\n" )
      if @notset;

    # parse extra and region parameters
    $self->pars( WECF::Pars->new( pars => $opt->par) );

    $self->_re( eval "qr{$opt->{extract_re}}" or die( "--extract_re: bad regular expression: ", $opt->extract_re ) );

    $self->_createdate( time2str( "%Y-%m-%dT%H:%M:%S", time, 'UTC' ) );

    # set date valid to creation date unless specified
    if ( defined $opt->vdate )
    {
	require Date::Manip;
	# make sure it's something valid
	my $date = Date::Manip::ParseDateString( $opt->vdate )
	  or die( "illegal date: ", $opt->vdate );

	$opt->{vdate} = Date::Manip::UnixDate($date, "%Y-%m-%dT%H:%M:%S" );
    }
    else
    {
	$opt->{vdate} = $self->createdate;
    }

    unless ( @ARGV )
    {
	@ARGV = <STDIN>;
	chomp @ARGV;
    }

    $self->files( \@ARGV );

    return;

}

1;
