#! perl

# --8<--8<--8<--8<--
#
# Copyright (C) 2011 Smithsonian Astrophysical Observatory
#
# This file is part of write-ecf-fits
#
# write-ecf-fits is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# -->8-->8-->8-->8--

use 5.10.0;

use strict;
use warnings;

package WECF::OutputParams;

our $VERSION = '2.0.0';

use Moo;

use Try::Tiny;

use PDL qw[ which ];
use PDL::Types;
use PDL::IO::RDB qw[ rdrdb ];
use Astro::FITS::CFITSIO::Simple qw[ rdfits ];
use Data::Alias;
use Params::Validate qw[ validate validate_with HASHREF SCALAR ];

use Util::Any -list => [ qw( mapp ) ];

use WECF::OutputParams::Group;

has file => ( is => 'ro',
	      required => 1 );

has filter => ( is => 'ro',
		predicate => 'has_filter'
	      );

has _default => ( is => 'rw',
		  init_arg => undef,
		);

has _group => ( is => 'rw',
		init_arg => undef,
	      );

has _parvals => ( is => 'ro',
		  init_arg => undef,
		  lazy => 1,
		  builder => '_build_parvals'
		);

# which points on the grid are requested for output
has pmap    => ( is => 'rw',
		 init_arg => undef,
		 lazy => 1,
		 builder => '_build_pmap'
	       );

# which points on the grid are linked to (origin of linked data)
has lmap    => ( is => 'rw',
		 init_arg => undef,
		 lazy => 1,
		 builder => '_build_lmap'
	       );


use Readonly;

Readonly my @par_names => qw[ energy theta phi ];
Readonly my %par_names => map { $_ => 1 } @par_names;

sub group {

    my ( $self, $group ) = @_;

    return $self->_group->{$group};
}

sub default { $_[0]->_default() }

sub names { @par_names }

sub _build_pmap {

    my $self = shift;

    return $self->mask( points => $self->default, index => 1 );
}

sub _build_lmap {

    my $self = shift;

    my $mask = $self->mask;

    $self->mask( points => $self->group( $_ )->data, mask => $mask )
      foreach keys %{ $self->_group};

    return $mask;
}

sub BUILD {

    my $self = shift;

    my $have_filter = $self->has_filter && defined $self->filter;

    my %pars;

    # Read in list of parameters
    given( $self->file ) {

	when( /\.fits$/ )  {
	    %pars = rdfits( $_, { defdtype => double } );
	}

	when( /\.rdb$/ ) {

	    %pars = rdrdb( $_ );

	}

	default {
	    die( "$_: unknown format for output params file",
		 "(bad or no filename suffix?)\n" );
	}
    }

    if ( $have_filter )
    {

	package WECF::OutputParams::Eval;
	use PDL;
	use Data::Alias;

	use strict;
	use warnings;

	my $eval = join( ";\n",
			 ( map { "alias my \$$_ = \$pars{$_}" }
			   keys %pars ),
			 $self->filter );
	my $select = eval $eval;
	die( $@ ) if $@;
	$_ = $_->index(which($select))->sever foreach values %pars;
    }

    # See if there were any extra (mapping) columns.
    # mapping columns match
    #  /^(?<group>[[:alnum:]]+)_(?<parameter>.*)$/
    # where each group must contain all of the parameters in
    # $self->par_names

    my $re = qr/^(?<group>[[:alnum:]]+)
		_
		(?<parameter>
		    ${ \(
		    join('|', @{ \@par_names })
		)
	    }
	)$
	       /x;


    # extract candidates
    my %group;
    for my $par ( keys %pars )
    {
	given ( $par ) {

	    # original parameter
	    when( \@par_names ) {
		$group{_default}{$_} = delete $pars{$_}
	    }
	    # group parameter
	    when( $re ) {
		$group{$+{group}}{$+{parameter}} = delete $pars{$_};
	    }
	}
    }

    die( $self->file, ': unrecognized column(s): ',
	 join( ', ', keys %pars ), "\n" )
      if %pars;


    try { $self->_validate_columns( $group{_default} ) }
      catch { chomp; die( "output parameters: $_\n" ) };

    # this needs to be set so that parvals can be built when
    # $self->idx is called by _validate_group
    $self->_default( delete $group{_default} );

    # create & store Group objects
    $self->_group ( { mapp
		    { $a => WECF::OutputParams::Group->new( name => $a,
							    data => $b,
							    pars => $self,
							  )
		  } %group
		    });

    return;
}

sub _validate_columns {

    my ( $self, $data ) = @_;

    my @bad = grep { ! exists $data->{$_} } @par_names;
    die( "missing columns: ", join( ', ', @bad ), "\n" )
      if @bad;

    return;
}



# created sorted piddles of unique parameter values
sub _build_parvals {

    my $self = shift;

    my %parvals = mapp { $a => $b->uniq } %{$self->default};

    return \%parvals;

}

sub parvals { return wantarray ? @{$_[0]->_parvals}{ @par_names }
		: $_[0]->_parvals;
	  }

# given index (into parvals) piddles for each parameter, return
# value piddles
sub vals {

    my $self = shift;
    my %index = 'HASH' eq ref $_[0] ? %{$_[0]} : @_;

    alias my %parvals = %{ $self->_parvals };
    my %value;

    while ( my ( $key, $val ) = each %index )
    {
	die( "unknown parameter name: $key\n" )
	  unless exists $parvals{$key};

	$value{$key} = $parvals{$key}->index( $val );

	# make the piddle non-point like
	$value{$key} = $value{$key}->dummy(0) unless $value{$key}->ndims;
    }

    return wantarray ? @value{ @par_names } : \%value;
}

# given value piddles for each parameter, return index (into
# parvals) piddles
sub idxs {

    my $self = shift;

    my %value = 'HASH' eq ref $_[0] ? %{$_[0]} : @_;

    alias my %parvals = %{ $self->_parvals };
    my %index;

    while ( my ( $key, $val ) = each %value )
    {
	die( "unknown parameter name: $key\n" )
	  unless exists $parvals{$key};

	$index{$key} = vsearch( $val, $parvals{$key} );

	# make the piddle non-point like
	$index{$key} = $index{$key}->dummy(0) unless $index{$key}->ndims;

	die( "bad values for $key requested in idxs\n" )
	  unless all( $val == $parvals{$key}->index($index{$key}) );
    }

    return wantarray ? @index{ @par_names } : \%index;
}

###########################################################################
#
#       $mask = $self->mask( points => \%points,
#                            index  => bool (optional),
#                            init   => num|piddle
#                            mask   => piddle
#                          );
#
#    Create an n-Dimensional mask indicating the required combinations of
#    independent parameters. Inputs are
#
#    %points
#        A hash of 1D-piddles, keyed off of parameter names. These are the
#        allowed values for the input data.
#
#    $index
#        if true, the mask is set to the index of the point in the
#        passed piddles
#
#    init
#        initialize the mask values to this before constructing it
#
#    mask
#        use the passed piddle to store the mask
#
#    mask returns an N-dimensional piddle ($mask) whose shape is
#    given by the length of @params and the size of the
#    corresponding piddles in "%points". Each element in $mask
#    represents a legal combination of parameter values. It is set
#    to 1 if that element is specified in %set, 0 if not.
#


sub mask
{
    my $self = shift;

    my %args = validate( @_,
		       {
			points => { type => HASHREF, optional => 1 },
			index  => { type => SCALAR, default => 0 },
			init   => { default => undef },
			mask   => { default => undef }
		       }
		       );

    alias my %parvals = %{$self->parvals};

    my $mask = $args{mask} //
      zeroes( $args{index} ? long : byte, map { $parvals{$_}->nelem } @par_names );

    $mask .= $args{init} if defined $args{init};

    if ( defined $args{points} )
    {
	my $idx = cat( $self->idxs( %{$args{points}} ) )->transpose;
	# arrgh. "feature" in Perl 5.10+ causes perl debugger to not
	# like lvalue .= for functions returning piddles
	my $tmp = $mask->range( $idx );
	$tmp .= $args{index} ? $tmp->sequence : 1;
    }

    return $mask;
}


# test if the passed parameter tuple is being linked to. this is indicated
# by the grid point being set in lmap;
sub is_linked_to {

    my $self = shift;

    my %args = validate( @_,
		       {
			points => { type => HASHREF, optional => 0 },
		       }
		       );

    # create indices into the mask for the passed points
    my $idx = PDL::cat( $self->idxs($args{points}) )->transpose;

    # and return the lmap values on the grid
    return $self->lmap->range( $idx );
}



1;
