#! perl

# --8<--8<--8<--8<--
#
# Copyright (C) 2011 Smithsonian Astrophysical Observatory
#
# This file is part of write-ecf-fits
#
# write-ecf-fits is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# -->8-->8-->8-->8--

use 5.10.0;

use strict;
use warnings;

package WECF::OutputParams::Group;

our $VERSION = '2.0.0';

use Moo;

use Try::Tiny;
use Params::Validate qw[ validate HASHREF ];
use PDL::Lite;

use Readonly;
Readonly my @group_names => qw[ _default stats ];

has name => ( is => 'ro',
	      isa => sub { die( "unrecognized group: $_[0]\n" )
			     unless $_[0] ~~ \@group_names },
	      required => 1 );
has data => ( is => 'ro', required => 1 );
has pmap  => ( is => 'rw',
	       init_arg => undef );
has pars => ( is => 'ro', required => 1,
	      isa => sub { die "pars is not an 'OutputParams'"
			     unless $_[0]->isa( 'WECF::OutputParams' )
			 },
	    );

has is_identity => ( is => 'ro',
		     init_arg => undef,
		     lazy => 1,
		     builder => '_is_identity'
		   );

sub BUILD {
    my $self = shift;

    my $pars = $self->pars;

    try { $pars->_validate_columns( $self->data ) }
      catch { chomp; die( $self->name, " parameter group: $_\n" ) };

    # generate index mask; if there's an error, the indices aren't in
    # the list of output parameter sets
    try {

	$self->pmap( $pars->mask( points => $self->data, index => 1 ) );

    } catch {

	given ($_ ) {
	    when( /bad values/ ) {
		die( $self->name,
		     " parameter group: some sets are outside of output parameter space\n" )
	    }
	    default {
		chomp;
		die( $self->name,
		     " parameter group: internal error: $_\n" );
	    }
	}

    }

}

sub _is_identity {

    my $self = shift;

    return ($self->pmap == $self->pars->pmap)->all;

}

# map a set of output parameter values onto the group parameter values
sub mapvals {

    my $self = shift;
    my %args = validate( @_,
		       {
			points => { type => HASHREF, optional => 0 },
			flatten => { default => 0 },
		       }
		       );

    return $args{points}
      if $self->is_identity;

    my $pars = $self->pars;

    # first extract indices from the OutputParams pmap attribute
    my $idx = PDL::cat( $pars->idxs($args{points}) )->transpose;
    my $idxs = $pars->pmap->range( $idx );

    # now grab the parameter values
    return map {
	my $val = $self->data->{$_}->index($idxs)->sever;
	$val = $val->sclr if $args{flatten};
	$_ => $val;
    } $pars->names;
}

1;
