#! perl

# --8<--8<--8<--8<--
#
# Copyright (C) 2011 Smithsonian Astrophysical Observatory
#
# This file is part of write-ecf-fits
#
# write-ecf-fits is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# -->8-->8-->8-->8--

use 5.10.0;

use strict;
use warnings;

package WECF::Par;

our $VERSION = '2.0.0';

use Moo;

has attrs => ( is => 'ro',
	       isa => sub { 'ARRAY' eq ref $_[0] },
	       required => 1
	     );

has cols => ( is => 'ro',
	      isa => sub { 'ARRAY' eq ref $_[0] },
	      required => 1
	    );

around cols => sub {
    my $orig = shift;
    my $self = shift;
    my $cols = $orig->( $self );
    return wantarray ? @$cols : $cols;
};


has $_ => ( is => 'ro',
	    required => 1 )
  for qw[ unit group type ];


package WECF::Pars;

use Moo;

use Tie::IxHash;
use Regexp::Common;
use Params::Validate qw[ validate ARRAYREF HASHREF SCALAR ];

use Util::Any -list => [ 'uniq' ];

use WECF::Datum;
use WECF::Datum::Bounds;
use WECF::Datum::Bounds::Abs;
use WECF::Datum::Bounds::Fraction;
use WECF::Datum::Bounds::Relative;

has _pars => (
	      is => 'rw',
	      init_arg => undef,
	     );

has cols => ( is => 'ro',
	      lazy => 1,
	      init_arg => undef,
	      builder => '_build_cols'
	    );

around cols => sub {
    my $orig = shift;
    my $self = shift;
    my $cols = $orig->( $self );
    return wantarray ? @$cols : $cols;
};

sub npar { shift()->_pars->Length }
sub par  { $_[0]->_pars->FETCH($_[1]) }
sub pars { shift()->_pars->Values }
sub fits_names { shift()->_pars->Keys }
sub fits_units { map { $_->{unit} } shift()->pars }

sub _build_cols {
    return [ uniq map { my $par = $_;
			map {
			    $par->attrs->{$_}
			}  $par->cols;
		    } shift()->pars
	   ] ;
}

# parse region and parameter specifications
#
# returns:
#   array of hashes with extra parameter info.
#   array of hashes with region parameter info.
#   array of FITS column names, including the confidence columns
#   array of FITS units,  including the confidence columns
#
# the input specs have the format
#
#	name,unit,col,cmin,cmax
#
# where
#
#   name = name of the column as it should appear in the FITS file
#   unit = units as they should appear in the FITS file
#   col  = the name of the column in the RDB table
#   cmin = the name of the column with the lower bound of the confidence region
#   cmax = the name of the column with the upper bound of the confidence region
#
#
# hash fields are:
#
#   name = name of the column as it should appear in the FITS file
#   unit = units as they should appear in the FITS file
#   col  = the name of the column in the RDB table
#   cmin = the name of the column with the lower bound of the confidence region
#   cmax = the name of the column with the upper bound of the confidence region
#   cmin_type = 'relative', 'absolute', or 'fraction'
#   cmax_type = 'relative', 'absolute', or 'fraction'

sub BUILD {

    my $self = shift;

    my %args = validate( @_, { pars => { type => ARRAYREF } } );

    my $pars = tie my %col, 'Tie::IxHash';

    foreach my $par ( @{ $args{pars} } )
    {
	my $nattr = my ( $name, $unit, $base, $cmin, $cmax ) = split(',', $par );

	given( $nattr ) {


	    when( 3 ) {

		$col{$name} = WECF::Par->new( type => 'WECF::Datum',
							     unit => $unit,
							     group => 'default',
							     cols => [ 'data' ],
							     attrs => { name => $name,
									data => $base
								      },
							   )
	    }
	    when( 5 ) {

		$col{$name} = WECF::Par->new( unit => $unit,
					     group => 'default',
					     type => 'WECF::Datum',
					     cols => [ 'data' ],
					     attrs => {
						       name => $name,
						       data => $base,
						      },
					   );

		my %par = ( cmin => $cmin, cmax => $cmax );

		for my $bnd ( qw/ cmin cmax / )
		{
		    my $col = $par{$bnd};
		    my $name = "${name}_s" . substr($bnd,1);

		    my %args = (
				unit => $unit,
				group => 'stats',
				cols => [ 'data', 'base' ],
			       );
		    my %attrs = ( name => $name,
				  base  => $base,
				);

		    given ($col) {

			# numeric value, fraction column
			when ( /$RE{num}{real}/ ) {

			    $col{$name} =
			      WECF::Par->new(
					    type => 'WECF::Datum::Bounds::Fraction',
					    %args,
					    cols => [ 'base' ],
					    attrs => {
						      %attrs,
						      data => $bnd eq 'cmin' ? 1 - $col : 1 + $col
						     }
					   );
			}

			# delta column
			when( /^[+]/ ) {
			    $col{$name} =
			      WECF::Par->new(
					    type => 'WECF::Datum::Bounds::Relative',
					    %args,
					    attrs => {
						      %attrs,
						      data => substr($col,1),
						      factor => $bnd eq 'cmin' ? -1 : 1
						     }
					   );
			}

			# absolute value
			default {
			    $col{$name} =
			      WECF::Par->new(
					    type => 'WECF::Datum::Bounds::Abs',
					    %args,
					    attrs => {
						      %attrs,
						      data => $col,
						     }
					   );
			}

		    }
		}
	    }


	    default {
		die( "illegal extra parameter specification: $par\n" );
	    }

	}
    }

    $self->_pars( $pars );
}

sub populate {

    my $self = shift;

    my %args = validate( @_,
		       {
			data => { type => HASHREF },
			links => { type => HASHREF },
			is_linked_to => { type => SCALAR },
		       } );

    my %data;
    for my $par ( $self->pars )
    {
	my %cpar = %{$par->attrs};

	$cpar{$_} = $args{data}->{$cpar{$_}}
	  for grep { defined $cpar{$_} } $par->cols;

	$data{$cpar{name}} = $par->type->new( %cpar,
					      link => $args{links}{$par->group} || {},
					      is_linked_to => $args{is_linked_to},
					    );
    }

    return %data;
}



1;

