#! perl

use strict;
use warnings;

use RDB;
use PDL::IO::Misc;
use PDL::IO::RDB qw[ rdrdb ];
use File::Spec::Functions;

use Test::Exception;

use Fennec::Lite plan => 36;
our ( $a, $b );

use Util::Any -list => [ qw ( mapp all )], { prefix => 1 };

use WECF::OutputParams;

use lib 'src';
require 'write-ecf-fits';

my $datadir = catdir( qw[ t data ] );


my @cols = qw[ theta energy phi ];

my %fid;
@fid{@cols} = rcols( catfile( $datadir, 'dr01.txt' ) );

tests 'formats' => sub {

    subtest 'rdb' => sub {

	my $pars = WECF::OutputParams->new( file => catfile( $datadir, 'dr01.rdb' ) );

	is_deeply( [ sort $pars->names ], [ sort @cols ], 'parameter names' );

	ok ( all($pars->default->{$_} == $fid{$_}), "rdb: $_" )
	  foreach $pars->names;
    };

      subtest 'fits' => sub {
	  my $pars = WECF::OutputParams->new( file => catfile( $datadir, 'dr01.fits' ) );

	  is_deeply( [ sort $pars->names ], [ sort @cols ], 'parameter names' );

	  ok ( all($pars->default->{$_} == $fid{$_}), "rdb: $_" )
	    foreach $pars->names;
      };

};

tests 'groups' => sub {

    my $pars = WECF::OutputParams->new( file => catfile( $datadir, 'dr02.rdb' ) );

    ok ( all($pars->default->{$_} == $fid{$_}), "groups, default: $_" )
      foreach $pars->names;
    ok ( all($pars->group('stats')->data->{$_} == $fid{$_}),
	 "groups, stats: $_" )
      foreach $pars->names;

    is( $pars->group('nop'), undef,  "unknown group requested" );

    throws_ok { WECF::OutputParams->new( file => catfile( $datadir, 'dr03.rdb' ) ) }
      qr/unrecognized group.*\sflat/, 'unrecognized group provided';

    throws_ok { WECF::OutputParams->new( file => catfile( $datadir, 'dr06.rdb' ) ) }
      qr/outside of output parameter space/, 'out of range group';

};


tests 'columns' => sub {

    throws_ok { WECF::OutputParams->new( file => catfile( $datadir, 'dr04.rdb' ) ) }
      qr/output parameters:\s+missing columns/, 'missing default column';

    throws_ok { WECF::OutputParams->new( file => catfile( $datadir, 'dr05.rdb' ) ) }
      qr/stats parameter group:\s+missing columns/, 'missing stats column';
};


tests 'values' => sub {

    my $pars = WECF::OutputParams->new( file => catfile( $datadir, 'dr01.rdb' ) );

    # reverse things to hopefully force bugs out
    my %sfid = list_mapp { $a => $b->uniq->slice('-1:0') } %fid;

    my %ns = list_mapp { $a => $b->sequence->slice('-1:0') } %sfid;

    my $idxs = $pars->idxs( %sfid );

    ok ( all( $idxs->{$_} == $ns{$_} ), "index $_" ) foreach keys %ns;

    # and go backwards

    my $vals = $pars->vals( %ns );
    ok ( all( $vals->{$_} == $sfid{$_} ), "value $_" ) foreach keys %sfid;

};

tests 'mask' => sub {


    my $pars = WECF::OutputParams->new( file => catfile( $datadir, 'dr07.rdb' ) );

    my $nexp = $pars->default->{energy}->nelem;

    subtest 'boolean mask' => sub {

	my $mask = $pars->mask( points => $pars->default );

	is( $mask->sum, $nexp, "element count" );
	my $idx = cat( $pars->idxs( $pars->default ) )->transpose;

	ok( all( $mask->range( $idx ) == 1 ),
	    'element values' );
    };

    subtest 'index mask' => sub {

	my $mask = $pars->mask( points => $pars->default,
				index => 1, init => -1 );

	is( ($mask != -1)->sum, $nexp, "element count" );
	my $idx = cat( $pars->idxs( $pars->default ) )->transpose;

	ok( all( $mask->range( $idx ) == $pars->default->{energy}->sequence ),
	    'element values' );
    };


};

tests 'map' => sub {

    subtest 'identical' => sub {
	my $pars = WECF::OutputParams->new( file => catfile( $datadir, 'dr02.rdb' ) );

	my $group = $pars->group('stats');

	is ( $group->is_identity, 1, 'stats identical' );

	my $mvals = $group->mapvals( points => $pars->default );

	ok( all( $mvals->{$_} == $pars->default->{$_} ), "mapped: $_" )
	  foreach $pars->names;

    };


    subtest 'mixed' => sub {


	my $if = catfile( $datadir, 'dr09.rdb' );
	my %fid = rdrdb( $if );

	my $pars = WECF::OutputParams->new( file => $if );

	my $group = $pars->group('stats');

	is ( $group->is_identity, 0, 'stats not identical' );

	# test piddles
        {
	    my %mapvals = $group->mapvals( points => $pars->default );

	    ok( all( $mapvals{$_} == $fid{'stats_' . $_} ), "mapval: pdl $_" )
	      foreach $pars->names;
	}

	# test scalars
        {
	    my %pars = %{ $pars->default };

	    for my $idx ( 0.. $pars{energy}->nelem-1 )
	    {
		my %point =
		  map {
		      $_ => $pars->default->{$_}->at($idx)
		  } $pars->names;

		my %mapvals = $group->mapvals( points => \%point );

		for my $name ( $pars->names )
		{
		    my $exp = $fid{"stats_$name"}->at($idx);
		    ok( $mapvals{$name}->sclr ==  $exp,
			"mapval: scalar $name = $exp" );
		}
	    }

	}
    };


    subtest 'linked' => sub {

	my $if = catfile( $datadir, 'dr08.rdb' );

	my %fid = rdrdb( $if );
	my $rdb = RDB->new( $if )
	  or die( "$if: error reading/parsing\n" );

	my $pars = WECF::OutputParams->new( file => $if );

	my $group = $pars->group('stats');

	is ( $group->is_identity, 0, 'stats not identical' );

	my %linked_to;
	my %row;
	my @rows;
	while( $rdb->read( \%row ) )
	{
	    # ensure that they're numbers;

	    $row{$_} += 0 for keys %row;
	    my %stat = map { (my $key = $_ ) =~ s/stats_//;
			     $key => delete $row{$_}
			 } grep { /stats_/ } keys %row;

	    # self referential links don't count
	    $linked_to{$stat{energy}}{$stat{theta}}{$stat{phi}} +=
	      ! list_all { $row{$_} == $stat{$_} } keys %row ;

	    push @rows, {%row};
	}

	my $linked_to =
	  pdl( map {
	              ( $linked_to{$_->{energy}}{$_->{theta}}{$_->{phi}} // 0 )== 1
		  } @rows
	     );

	delete $fid{$_} for grep { /stats_/ } keys %fid;

	ok( all( $linked_to == $pars->is_linked_to( points => \%fid )),
	    'linked: linked_to')


    };


};

tests 'filter' => sub {

    my %mask = ( inc => sub { which( $fid{energy} == 1.0 ) },
		 exc => sub { which( $fid{energy} != 1.0 ) },
	       );

    my %tpar = ( 'dr01.fits' => { inc => '$energy == 1.0',
				  exc => '$energy != 1.0'
				},
		 'dr01.rdb'  => { inc => '$energy == 1.0',
				  exc => '$energy != 1.0'
				},
	       );

    while( my ($file, $filter) = each %tpar )
    {
	( my $ext ) = $file =~ /(fits|rdb)$/;

	for my $filt ( qw[ inc exc ] )
	{
	    my $opars = WECF::OutputParams->new( file => catfile( $datadir, $file ), filter => $filter->{$filt});
	    my $pars = $opars->default;

	    my $mask = $mask{$filt}->();
	    my %filtered = map { $_ => $fid{$_}->index($mask)->sever } keys %fid;
	    ok( all ( $pars->{$_} == $filtered{$_} ), "filter $ext: $filt ($_)" )
	      foreach keys %filtered;
	}

    }


};

run_tests;
