#! !perl

use strict;
use warnings;

use File::Find;
use File::Temp;

use PDL::IO::Misc;
use PDL::IO::RDB qw[ rdrdb ];
use File::Spec::Functions;
use Test::Exception;

use Util::Any -list => [ qw( mapp ) ];


use Fennec::Lite plan => 107;

use RDB;

use WECF::Options;

use lib 'src';
require 'write-ecf-fits';

use lib 't';
require 'common.pl';

use vars qw[ $opt ];

my $datadir = catdir( qw[ t data ] );
my @params = qw [ x y sma smb pa ];

my $idxfile = catfile( $datadir, 'dr01.rdb');

my @files;
find( sub { push @files, $File::Find::name if $_ eq 'ell.rdb' },
      catdir( $datadir, 'ecf' ) );

my @argv = ( qw[ --output foo.fits
		 --detector hrc-s
		 --config none
		 --frac ecf
	      ],
	   );
my @pars = map { ( '--par' => "$_,${_}_units,$_" ) } @params;

tests 'all ok' => sub {

    @ARGV = ( @argv, @pars, '--output_params', $idxfile, @files );

    $opt = WECF::Options->new;

    lives_ok { read_data() } 'all data accounted for';
};

# remove one of the input data files (and thus not fulfilling the required
# set) and see what happens
tests 'missing' => sub {

    my @message;
    local $SIG{__WARN__} = sub { push @message, @_ };

    my @files = @files;

    pop @files;
    @ARGV = ( @argv, @pars, '--output_params', $idxfile, @files );

    $opt = WECF::Options->new;

    throws_ok { read_data() } qr/errors in processing/i, 'missing data: exception';

    my $err = join( "\n", @message );
    like ( $err, qr/^\s*1\s+2\s+329\s*$/m, 'missing data: warnings' );
};

# remove one of the required sets of parameters from the required
# set and see what happens
tests 'extra' => sub {

    my @message;
    local $SIG{__WARN__} = sub { push @message, @_ };

    my @files = @files;

    my $tmp = File::Temp->new( SUFFIX => '.rdb' );

    my $in = RDB->new( $idxfile, '<' )
      or die( "error opening/reading $idxfile\n" );

    my $out = RDB->new( $tmp->filename, '>' )
      or die( "error opening $tmp\n" );
    $out->init( $in );

    my %row;
    while( $in->read(\%row) ) {  $out->write( \%row ) if $row{phi} != 329 } ;
    $out->close;

    subtest 'warn on extra', sub {
	@ARGV = ( @argv, @pars, '--output_params', $tmp->filename, @files );

	$opt = WECF::Options->new;

	lives_ok( sub { read_data() }, 'extra data: no exception' )
	  or BAIL_OUT( 'cannot continue after error' );

	my $err = join( "\n", @message );
	like ( $err, qr/extra theta/m, 'extra data: warnings' );
    };

    subtest 'error on extra', sub {
	@ARGV = ( @argv, @pars, '--err_extra_params', '--output_params', $tmp->filename, @files );

	$opt = WECF::Options->new;

	throws_ok( sub { read_data() }, qr/errors in processing/i, 'exception' )
	  or BAIL_OUT( 'cannot continue after error' );

	my $err = join( "\n", @message );
	like ( $err, qr/extra theta/m, 'extra data: warnings' );
    };


};

# chck that an error is thrown if multiple input files have the same independent parameters
tests 'duplicate' => sub {

    my @message;
    local $SIG{__WARN__} = sub { push @message, @_ };

    my @files = @files;

    push @files, $files[0];

    @ARGV = ( @argv, @pars, '--output_params', $idxfile, @files );

    $opt = WECF::Options->new;

    $files[0] =~ $opt->re;
    my %param = map { $_ => $+{$_} + 0.0 } keys %+;
    throws_ok { read_data() } qr/errors in processing/i, 'duplicate data: exception';

    my $err = join( "\n", @message );
    my $exp =  join( '', "duplicate data for ", join( ', ', map { "$_ = $param{$_}" } qw[ energy theta phi ] ) );
    like ( $err, qr/^$exp/m, 'duplicate data: warnings' );
};

tests 'pmap' => sub {


    my @files;
    find( sub { push @files, $File::Find::name if $_ eq 'ell_abs.rdb' },
	  catdir( $datadir, 'ecf' ) );

    my $idxfile = catfile( $datadir, 'dr09.rdb');

    @ARGV = ( @argv,
	      ( map { ( '--par' => "$_,${_}_units,$_,${_}_smin,${_}_smax" ) } @params ),
	      '--output_params', $idxfile,
	      @files
	    );

    $opt = WECF::Options->new;

    my ( $fraction, $pars, $data ) = read_data();

    my $rdb = RDB->new( $idxfile )
      or die( "error reading $idxfile\n" );

    my %pset;
    while( $rdb->read( \%pset  ) )
    {
	no autovivification 'strict';

	# check that primary columns are identical
	my $pfile = catfile( $datadir, 'ecf',
			     @pset{qw[ energy theta phi ]},
			     'ell_abs.rdb' );

	my $lfile = catfile( $datadir, 'ecf',
			     @pset{qw[ stats_energy stats_theta stats_phi ]},
			     'ell_abs.rdb' );


	my $pfx = join('; ',
		       join( ', ',  @pset{ qw[ energy theta phi ]}),
		       join( ', ',  @pset{ map { "stats_$_" } qw[ energy theta phi ]} )
		       );

	my %ecf = read_fiducial( $pfile );
	my %lecf = read_fiducial( $lfile );

	$pset{$_} += 0 for keys %pset;
	my %lset = map { (my $key = $_ ) =~ s/stats_//;
			 $key => delete $pset{$_}
		     } grep { /stats_/ } keys %pset;


	for my $par ( @params )
	{
	    # extract primary and stats (linked)
	    my $pdata = $data->get( %pset, name => $par, raise => 1 );
	    my $ldata = $data->get( %lset, name => $par, raise => 1 );


	    ok( all ( $ecf{$par}  == $pdata->value ), "ecf: $pfx: $par" );
	    ok( all ( $lecf{$par} == $ldata->value ), "lecf:$pfx: $par" );

	    # make sure primary is not linked
	    ok( all ( $ecf{$par} == $data->resolve( %pset, name => $par ) ),
		"ecf resolve: $pfx: $par" );

	    # make sure links work
	    for my $sfx ( qw[ smin smax ] )
	    {
		my $spar = "${par}_${sfx}";
		my $exp = $ecf{$par} + $lecf{$spar} - $lecf{$par};

		ok( all ( approx( $exp,
				  $data->resolve( %pset, name => $spar )
				)
			),
		    "lecf <=> pdata: $pfx: $spar" );
	    }
	}


    }

};

run_tests;
